@@ -18,23 +18,21 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
-import os
-import sys
-from bzrlib import errors, ui
-from bzrlib.option import Option
+from bzrlib import errors
 from bzrlib.commands import Command
+from bzrlib.option import Option
 import bzrlib.builtins
 
 from bzrlib.lazy_import import lazy_import
 lazy_import(globals(), '''
-import signal, shlex, thread
+import sys
 
-from PyQt4 import QtGui, QtCore
+from PyQt4 import QtCore, QtGui
 
 from bzrlib import (
     builtins,
-    commands,
     osutils,
+    ui,
     )
 from bzrlib.branch import Branch
 from bzrlib.bzrdir import BzrDir
@@ -43,6 +41,7 @@ from bzrlib.workingtree import WorkingTree
 from bzrlib.plugins.qbzr.lib import i18n
 from bzrlib.plugins.qbzr.lib.add import AddWindow
 from bzrlib.plugins.qbzr.lib.annotate import AnnotateWindow
+from bzrlib.plugins.qbzr.lib.branch import QBzrBranchWindow
 from bzrlib.plugins.qbzr.lib.browse import BrowseWindow
 from bzrlib.plugins.qbzr.lib.cat import (
     QBzrCatWindow,
@@ -61,13 +60,13 @@ from bzrlib.plugins.qbzr.lib.main import QBzrMainWindow
 from bzrlib.plugins.qbzr.lib.pull import (
     QBzrPullWindow,
     QBzrPushWindow,
-    QBzrBranchWindow,
     QBzrMergeWindow,
     )
 from bzrlib.plugins.qbzr.lib.revert import RevertWindow
-from bzrlib.plugins.qbzr.lib.subprocess import SubprocessUIFactory
 from bzrlib.plugins.qbzr.lib.tag import TagWindow
+from bzrlib.plugins.qbzr.lib.tree_branch import TreeBranch
 from bzrlib.plugins.qbzr.lib.uncommit import QBzrUncommitWindow
+from bzrlib.plugins.qbzr.lib.update import QBzrUpdateWindow
 from bzrlib.plugins.qbzr.lib.util import (
     FilterOptions,
     is_valid_encoding,
@@ -77,7 +76,6 @@ from bzrlib.plugins.qbzr.lib.uifactory import QUIFactory
 from bzrlib.plugins.qbzr.lib.send import SendWindow
 ''')
 
-from bzrlib.plugins.qbzr.lib import MS_WINDOWS
 from bzrlib.plugins.qbzr.lib.diff_arg import DiffArgProvider
 
 CUR_DIR=u'.'
@@ -155,23 +153,33 @@ class QBzrCommand(Command):
         # opening a branch, etc will go to the ui_factory to request a
         # password, and the app must exist before the dialog can be created.
         self._application = QtGui.QApplication(sys.argv)
-        ui.ui_factory = QUIFactory()
-        
-        # Set up global exception handling.
-        from bzrlib.plugins.qbzr.lib.trace import excepthook
-        sys.excepthook = excepthook
-        
+        std_ui_factory = ui.ui_factory
         try:
-            ret_code = self._qbzr_run(*args, **kwargs)
-            if ret_code is None:
-                main_window = getattr(self, "main_window", None)
-                if main_window is not None:
-                    ret_code = getattr(main_window, "return_code", None)
-            return ret_code
-        except Exception:
-            ui_mode = ("ui_mode" in kwargs and kwargs["ui_mode"])
-            from bzrlib.plugins.qbzr.lib.trace import report_exception
-            return report_exception(ui_mode=ui_mode)
+            ui.ui_factory = QUIFactory()
+            
+            # Set up global exception handling.
+            from bzrlib.plugins.qbzr.lib.trace import excepthook
+            sys.excepthook = excepthook
+            
+            try:
+                try:
+                    ret_code = self._qbzr_run(*args, **kwargs)
+                finally:
+                    # ensure we flush clipboard data (see bug #503401)
+                    clipboard = self._application.clipboard()
+                    clipEvent = QtCore.QEvent(QtCore.QEvent.Clipboard)
+                    self._application.sendEvent(clipboard, clipEvent)
+                if ret_code is None:
+                    main_window = getattr(self, "main_window", None)
+                    if main_window is not None:
+                        ret_code = getattr(main_window, "return_code", None)
+                return ret_code
+            except Exception:
+                ui_mode = kwargs.get("ui_mode", False)
+                from bzrlib.plugins.qbzr.lib.trace import report_exception
+                return report_exception(ui_mode=ui_mode)
+        finally:
+            ui.ui_factory = std_ui_factory
 
 
 ui_mode_option = Option("ui-mode", help="Causes dialogs to wait after the operation is complete.")
@@ -371,46 +379,60 @@ class cmd_qdiff(QBzrCommand, DiffArgProvider):
     aliases = ['qdi']
 
     def get_diff_window_args(self, processEvents):
-        from bzrlib.diff import _get_trees_to_diff
+        try:
+            from bzrlib.diff import get_trees_and_branches_to_diff
+            has_get_trees_and_branches_to_diff = True
+        except ImportError:
+            from bzrlib.diff import _get_trees_to_diff
+            has_get_trees_and_branches_to_diff = False
+        
+        if has_get_trees_and_branches_to_diff:
+            (old_tree, new_tree,
+             old_branch, new_branch,
+             specific_files, extra_trees) = \
+               get_trees_and_branches_to_diff(self.file_list, self.revision,
+                                              self.old, self.new)
+        else:
+            old_tree, new_tree, specific_files, extra_trees = \
+                    _get_trees_to_diff(self.file_list, self.revision,
+                                       self.old, self.new)
+            processEvents()
+            
+            if self.file_list:
+                default_location = self.file_list[0]
+            else:
+                # If no path is given, the current working tree is used
+                default_location = CUR_DIR
 
-        old_tree, new_tree, specific_files, extra_trees = \
-                _get_trees_to_diff(self.file_list, self.revision,
-                                   self.old, self.new)
-        processEvents()
-        
-        if self.file_list:
-            default_location = self.file_list[0]
-        else:
-            # If no path is given, the current working tree is used
-            default_location = CUR_DIR
-        
-        if self.old is None:
-            self.old = default_location
-        wt, old_branch, rp = \
-            BzrDir.open_containing_tree_or_branch(self.old)
-        processEvents()
-        if self.new is None:
-            self.new = default_location
-        if self.new != self.old :
-            wt, new_branch, rp = \
-                BzrDir.open_containing_tree_or_branch(self.new)
-        else:
-            new_branch = old_branch
-        processEvents()
-        
+            self_old = self.old
+            if self_old is None:
+                self_old = default_location
+            wt, old_branch, rp = \
+                BzrDir.open_containing_tree_or_branch(self_old)
+            processEvents()
+            self_new = self.new
+            if self_new is None:
+                self_new = default_location
+            if self_new != self_old :
+                wt, new_branch, rp = \
+                    BzrDir.open_containing_tree_or_branch(self_new)
+            else:
+                new_branch = old_branch
+            processEvents()
+            
         return old_tree, new_tree, old_branch, new_branch, specific_files
     
     def get_ext_diff_args(self, processEvents):
         args = []
         if self.revision and len(self.revision) == 1:
-            args.append("-r %s" % (self.revision[0].user_spec,))
+            args.append("-r%s" % (self.revision[0].user_spec,))
         elif self.revision and  len(self.revision) == 2:
-            args.append("-r %s..%s" % (self.revision[0].user_spec,
+            args.append("-r%s..%s" % (self.revision[0].user_spec,
                                        self.revision[1].user_spec))
-        
-        if self.new and not self.new==CUR_DIR:
+
+        if self.new and not self.new == CUR_DIR:
             args.append("--new=%s" % self.new)
-        if self.old and not self.old==CUR_DIR:
+        if self.old and not self.old == CUR_DIR:
             args.append("--old=%s" % self.old)
         
         if self.file_list:
@@ -628,12 +650,17 @@ class cmd_qbranch(QBzrCommand):
 
     takes_options = [simple_revision_option,
                      ui_mode_option]
+    try:
+        takes_options.append(bzr_option("branch", "bind"))
+    except KeyError:
+        # older version of bzr that doesn't support the option
+        pass
     takes_args = ['from_location?', 'to_location?']
 
     def _qbzr_run(self, from_location=None, to_location=None,
-                  revision=None, ui_mode=False):
+                  revision=None, bind=False, ui_mode=False):
         self.main_window = QBzrBranchWindow(from_location, to_location,
-                                  revision=revision, ui_mode=ui_mode)
+            revision=revision, bind=bind, ui_mode=ui_mode)
         self.main_window.show()
         self._application.exec_()
 
@@ -717,7 +744,6 @@ class cmd_qmain(QBzrCommand):
 
     Not finished -- DON'T USE
     """
-
     takes_options = []
     takes_args = []
     hidden = True
@@ -739,45 +765,27 @@ class cmd_qmain(QBzrCommand):
         self._application.exec_()
 
 
-class cmd_qsubprocess(Command):
+# [bialix 2010/02/04] body of cmd_qsubprocess has moved to subprocess.py
+# to see annotation of cmd_qsubprocess before move use:
+#     bzr qannotate commands.py -r1117
 
+class cmd_qsubprocess(Command):
+    """Run some bzr command as subprocess. 
+    Used with most of subprocess-based dialogs of QBzr.
+    
+    If CMD argument starts with @ characters then it used as name of file with
+    actual cmd string (in utf-8).
+    
+    With --bencode option cmd string interpreted as bencoded list of utf-8
+    strings. This is the recommended way to launch qsubprocess.
+    """
     takes_args = ['cmd']
+    takes_options = [Option("bencoded", help="Pass command as bencoded string.")]
     hidden = True
 
-    if MS_WINDOWS:
-        def __win32_ctrl_c(self):
-            import win32event
-            from bzrlib.plugins.qbzr.lib.subprocess import get_event_name
-            ev = win32event.CreateEvent(None, 0, 0, get_event_name(os.getpid()))
-            try:
-                win32event.WaitForSingleObject(ev, win32event.INFINITE)
-            finally:
-                ev.Close()
-            thread.interrupt_main()
-
-    def run(self, cmd):
-        if MS_WINDOWS:
-            thread.start_new_thread(self.__win32_ctrl_c, ())
-        else:
-            signal.signal(signal.SIGINT, sigabrt_handler)
-        ui.ui_factory = SubprocessUIFactory(sys.stdin, sys.stdout, sys.stderr)
-        if cmd.startswith('@'):
-            fname = cmd[1:]
-            f = open(fname, 'rb')
-            try:
-                cmd_utf8 = f.read()
-            finally:
-                f.close()
-        else:
-            cmd_utf8 = cmd.encode('utf8')
-        # XXX use bencode to avoid shlex, because it does not like backslashes
-        #     and therefore problematic on Windows.
-        argv = [p.decode('utf8') for p in shlex.split(cmd_utf8)]
-        commands.run_bzr(argv)
-
-
-def sigabrt_handler(signum, frame):
-    raise KeyboardInterrupt()
+    def run(self, cmd, bencoded=False):
+        from bzrlib.plugins.qbzr.lib.subprocess import run_subprocess_command
+        return run_subprocess_command(cmd, bencoded)
 
 
 class cmd_qgetupdates(QBzrCommand):
@@ -789,10 +797,15 @@ class cmd_qgetupdates(QBzrCommand):
 
     def _qbzr_run(self, location=CUR_DIR, ui_mode=False):
         branch, relpath = Branch.open_containing(location)
-        if branch.get_bound_location():
-            window = UpdateCheckoutWindow(branch, ui_mode=ui_mode)
+        tb = TreeBranch.open_containing(location, ui_mode=ui_mode)
+        if tb is None:
+            return errors.EXIT_ERROR
+        if tb.is_light_co():
+            window = QBzrUpdateWindow(tb.tree, ui_mode)
+        elif tb.is_bound():
+            window = UpdateCheckoutWindow(tb.branch, ui_mode=ui_mode)
         else:
-            window = UpdateBranchWindow(branch, ui_mode=ui_mode)
+            window = UpdateBranchWindow(tb.branch, ui_mode=ui_mode)
         self.main_window = window
         self.main_window.show()
         self._application.exec_()
@@ -821,7 +834,7 @@ class cmd_qhelp(QBzrCommand):
     hidden = True
 
     def _qbzr_run(self, topic):
-        window = show_help(topic)
+        show_help(topic)
         self._application.exec_()
 
 
@@ -912,7 +925,6 @@ class cmd_qupdate(QBzrCommand):
     takes_options = [ui_mode_option]
 
     def _qbzr_run(self, directory=None, ui_mode=False):
-        from bzrlib.plugins.qbzr.lib.update import QBzrUpdateWindow
         tree = open_tree(directory, ui_mode)
         if tree is None:
             return
@@ -1007,3 +1019,48 @@ class cmd_qbind(QBzrCommand):
         self.main_window = QBzrBindDialog(branch, location, ui_mode)
         self.main_window.show()
         self._application.exec_()
+
+
+class cmd_qrun(QBzrCommand):
+    """Run arbitrary bzr command.
+
+    If you wish to pass options to COMMAND, use ``--`` beforehand
+    so that the options aren't treated as options to the qrun
+    command itself. For example::
+
+      bzr qrun shelve -- --list
+
+    NOTE: you should use only canonical name of the COMMAND, not the alias.
+    """
+    takes_args = ['command?', 'parameters*']
+    takes_options = [ui_mode_option,
+        Option('directory',
+            help='Working directory.',
+            short_name='d',
+            type=unicode,
+            ),
+        Option('category',
+            help='Initial category selection.',
+            type=unicode,
+            ),
+        Option('execute',
+            help="Validate and run the supplied command immediately.",
+            short_name='e'
+            ),
+        ]
+    aliases = ['qcmd']
+
+    def _qbzr_run(self, command=None, parameters_list=None, ui_mode=False,
+        directory=None, category=None, execute=False):
+        from bzrlib.plugins.qbzr.lib.run import QBzrRunDialog
+        if parameters_list:
+            parameters = " ".join(parameters_list)
+        else:
+            parameters = None
+        if not command:
+            execute = False
+        window = QBzrRunDialog(command=command, parameters=parameters,
+            workdir=directory, category=category, ui_mode=ui_mode,
+            execute=execute)
+        window.show()
+        self._application.exec_()
