@@ -22,6 +22,7 @@
 import os.path
 from PyQt4 import QtCore, QtGui
 
+from bzrlib import errors
 from bzrlib.plugins.qbzr.lib.spellcheck import SpellCheckHighlighter, SpellChecker
 from bzrlib.plugins.qbzr.lib.autocomplete import get_wordlist_builder
 from bzrlib.plugins.qbzr.lib.commit_data import QBzrCommitData
@@ -44,7 +45,6 @@ from bzrlib.plugins.qbzr.lib.util import (
 
 from bzrlib.plugins.qbzr.lib.logwidget import LogList
 from bzrlib.plugins.qbzr.lib import logmodel
-from bzrlib.plugins.qbzr.lib.trace import *
 from bzrlib.plugins.qbzr.lib.uifactory import ui_current_widget
 from bzrlib.plugins.qbzr.lib.treewidget import (
     TreeWidget,
@@ -176,7 +176,7 @@ class PendingMergesList(LogList):
 
     def create_context_menu(self):
         super(PendingMergesList, self).create_context_menu()
-        showinfo = QtGui.QAction("Show &infomation...", self)
+        showinfo = QtGui.QAction("Show &information...", self)
         self.context_menu.insertAction(self.context_menu.actions()[0],
                                        showinfo)
         self.context_menu.setDefaultAction(showinfo)
@@ -193,11 +193,12 @@ class PendingMergesList(LogList):
         if index is None:
             index = self.currentIndex()
         
+        # XXX We should make this show all selected revsions...
+        
         revid = str(index.data(logmodel.RevIdRole).toString())
         branch = self.graph_provider.get_revid_branch(revid)
-        rev = self.graph_provider.load_revisions([revid])[revid]
         parent_window = self.window()
-        window = RevisionView(rev, branch, parent=parent_window)
+        window = RevisionView(revid, branch, parent=parent_window)
         window.show()
         parent_window.windows.append(window)
     
@@ -281,9 +282,7 @@ class CommitWindow(SubProcessDialog):
 
         self.filelist = TreeWidget(self)
         self.filelist.throbber = self.throbber
-        self.filelist.tree_model.is_item_in_select_all = lambda item: (
-            item.change is not None and
-            item.change.is_versioned())
+        self.filelist.tree_model.set_select_all_kind('versioned')
         
         self.file_words = {}
         self.connect(self.filelist.tree_model,
@@ -305,8 +304,7 @@ class CommitWindow(SubProcessDialog):
         self.message.setCompleter(self.completer)
         self.message.setAcceptRichText(False)
 
-        spell_highlighter = SpellCheckHighlighter(self.message.document(),
-                                                  spell_checker)
+        SpellCheckHighlighter(self.message.document(), spell_checker)
 
         grid.addWidget(self.message, 0, 0, 1, 2)
 
@@ -437,19 +435,36 @@ class CommitWindow(SubProcessDialog):
                     self.pending_merges_list.load_branch(self.tree.branch,
                                                          None,
                                                          self.tree)
+                    # Force the loading of the revisions, before we start
+                    # loading the file list.
+                    self.pending_merges_list._load_visible_revisions()
                     self.processEvents()
                 
                 self.filelist.tree_model.checkable = not self.pending_merges_list
                 self.is_loading = True
+                # XXX Would be nice if we could only load the files when the
+                # user clicks on the changes tab, but that would mean that
+                # we can't load the words list.
                 if not refresh:
                     fmodel = self.filelist.tree_filter_model
-                    #fmodel.setFilter(fmodel.UNVERSIONED, False)
+                    
+                    want_unversioned = self.show_nonversioned_checkbox.isChecked()
+                    fmodel.setFilter(fmodel.UNVERSIONED, want_unversioned)
+                    if not want_unversioned and self.initial_selected_list:
+                        # if there are any paths from the command line that
+                        # are not versioned, we want_unversioned.
+                        for path in self.initial_selected_list:
+                            if not self.tree.path2id(path):
+                                want_unversioned = True
+                                break
+                    
                     self.filelist.set_tree(
                         self.tree,
                         branch=self.tree.branch,
                         changes_mode=True,
-                        want_unversioned=self.show_nonversioned_checkbox.isChecked(),
-                        initial_checked_paths=self.initial_selected_list)
+                        want_unversioned=want_unversioned,
+                        initial_checked_paths=self.initial_selected_list,
+                        change_load_filter=lambda c:not c.is_ignored())
                 else:
                     self.filelist.refresh()
                 self.is_loading = False
@@ -460,7 +475,7 @@ class CommitWindow(SubProcessDialog):
         finally:
             self.throbber.hide()
             self.refresh_button.setDisabled(False)
-
+    
     def refresh(self):
         self.load(True)
 
@@ -483,7 +498,7 @@ class CommitWindow(SubProcessDialog):
                     file_words.add(os.path.split(path)[-1])
                     change = self.filelist.tree_model.inventory_data_by_path[
                                                                ref.path].change
-                    if change.is_renamed():
+                    if change and change.is_renamed():
                         file_words.add(change.oldpath())
                         file_words.add(os.path.split(change.oldpath())[-1])
                     #if num_versioned_files < MAX_AUTOCOMPLETE_FILES:
@@ -555,7 +570,7 @@ class CommitWindow(SubProcessDialog):
             ci_data.save()
 
     def wipe_commit_data(self):
-        if (self.tree.branch.control_files.get_physical_lock_status()
+        if (self.tree.branch.get_physical_lock_status()
             or self.tree.branch.is_locked()):
             # XXX maybe show this in a GUI MessageBox (information box)???
             from bzrlib.trace import warning
@@ -566,12 +581,13 @@ class CommitWindow(SubProcessDialog):
     def do_start(self):
         args = ["commit"]
         files_to_add = ["add", "--no-recurse"]
+        add_cmd_len = len(files_to_add)
         
         message = unicode(self.message.toPlainText()).strip() 
         if not message: 
             QtGui.QMessageBox.warning(self,
                 "QBzr - " + gettext("Commit"),
-                gettext("You should provide commit message."),
+                gettext("You should provide a commit message."),
                 gettext('&OK'))
             # don't commit, but don't close the window either
             self.on_failed()
@@ -634,7 +650,7 @@ class CommitWindow(SubProcessDialog):
         
         dir = self.tree.basedir
         commands = []
-        if len(files_to_add)>1:
+        if len(files_to_add) > add_cmd_len:
             commands.append((dir, files_to_add))
         commands.append((dir, args))
 
@@ -645,10 +661,16 @@ class CommitWindow(SubProcessDialog):
         """Show/hide non-versioned files."""
         if state and not self.filelist.want_unversioned:
             state = self.filelist.get_state()
-            self.filelist.set_tree(self.tree, changes_mode=True,
-                                   want_unversioned=True)
+            self.filelist.set_tree(
+                self.tree, changes_mode=True, want_unversioned=True,
+                change_load_filter=lambda c:not c.is_ignored())
             self.filelist.restore_state(state)
         
+        if state:
+            self.filelist.tree_model.set_select_all_kind('all')
+        else:
+            self.filelist.tree_model.set_select_all_kind('versioned')
+        
         fmodel = self.filelist.tree_filter_model
         fmodel.setFilter(fmodel.UNVERSIONED, state)
 
@@ -691,28 +713,34 @@ class CommitWindow(SubProcessDialog):
         @param  dialog_action:  purpose of parent window (main action)
         """
         # XXX make this function universal for both qcommit and qrevert (?)
-        checked = []        # checked versioned
-        unversioned = []    # checked unversioned (supposed to be added)
-        for ref in self.filelist.tree_model.iter_checked():
-            if ref.file_id:
-                checked.append(ref.path)
+        if self.filelist.tree_model.checkable:
+            checked = []        # checked versioned
+            unversioned = []    # checked unversioned (supposed to be added)
+            for ref in self.filelist.tree_model.iter_checked():
+                if ref.file_id:
+                    checked.append(ref.path)
+                else:
+                    unversioned.append(ref.path)
+    
+            if checked:
+                arg_provider = InternalWTDiffArgProvider(
+                    self.tree.basis_tree().get_revision_id(), self.tree,
+                    self.tree.branch, self.tree.branch,
+                    specific_files=checked)
+                
+                show_diff(arg_provider, ext_diff=ext_diff, parent_window = self)
             else:
-                unversioned.append(ref.path)
-
-        if checked:
+                msg = "No changes selected to " + dialog_action
+                QtGui.QMessageBox.warning(self,
+                    "QBzr - " + gettext("Diff"),
+                    gettext(msg),
+                    QtGui.QMessageBox.Ok)
+    
+            if unversioned:
+                # XXX show infobox with message that not all files shown in diff
+                pass
+        else:
             arg_provider = InternalWTDiffArgProvider(
                 self.tree.basis_tree().get_revision_id(), self.tree,
-                self.tree.branch, self.tree.branch,
-                specific_files=checked)
-            
+                self.tree.branch, self.tree.branch)            
             show_diff(arg_provider, ext_diff=ext_diff, parent_window = self)
-        else:
-            msg = "No changes selected to " + dialog_action
-            QtGui.QMessageBox.warning(self,
-                "QBzr - " + gettext("Diff"),
-                gettext(msg),
-                QtGui.QMessageBox.Ok)
-
-        if unversioned:
-            # XXX show infobox with message that not all files shown in diff
-            pass
