@@ -90,10 +90,17 @@ class QBzrConfigWindow(QBzrDialog):
         generalGrid.addWidget(self.emailEdit, 1, 1)
 
         self.editorEdit = QtGui.QLineEdit()
+        btnEditorBrowse = QtGui.QPushButton(gettext('&Browse...'))
+        self.connect(btnEditorBrowse,
+            QtCore.SIGNAL("clicked()"),
+            self.browseEditor)
+        editorHBox = QtGui.QHBoxLayout()
+        editorHBox.addWidget(self.editorEdit)
+        editorHBox.addWidget(btnEditorBrowse)
         label = QtGui.QLabel(gettext("&Editor:"))
         label.setBuddy(self.editorEdit)
         generalGrid.addWidget(label, 2, 0)
-        generalGrid.addWidget(self.editorEdit, 2, 1)
+        generalGrid.addLayout(editorHBox, 2, 1)
 
         self.emailClientCombo = QtGui.QComboBox()
         for name, label in _mail_clients:
@@ -374,17 +381,10 @@ class QBzrConfigWindow(QBzrDialog):
         qconfig = QBzrGlobalConfig()
         qparser = qconfig._get_parser()
 
-        if 'DEFAULT' not in parser:
-            parser['DEFAULT'] = {}
-
-        # Name & e-mail
-        username = '%s <%s>' % (
-            unicode(self.nameEdit.text()),
-            unicode(self.emailEdit.text()))
-        parser['DEFAULT']['email'] = username
-
         def set_or_delete_option(parser, name, value):
             if value:
+                if 'DEFAULT' not in parser:
+                    parser['DEFAULT'] = {}
                 parser['DEFAULT'][name] = value
             else:
                 try:
@@ -392,6 +392,12 @@ class QBzrConfigWindow(QBzrDialog):
                 except KeyError:
                     pass
 
+        # Name & e-mail
+        username = '%s <%s>' % (
+            unicode(self.nameEdit.text()),
+            unicode(self.emailEdit.text()))
+        set_or_delete_option(parser, 'email', username)
+
         # Editor
         editor = unicode(self.editorEdit.text())
         set_or_delete_option(parser, 'editor', editor)
@@ -574,3 +580,10 @@ class QBzrConfigWindow(QBzrDialog):
                         item.setCheckState(0, QtCore.Qt.Unchecked)
                 changed_item.setCheckState(0, QtCore.Qt.Checked)
                 self.extMergeListIgnore = False
+
+    def browseEditor(self):
+        filename = QtGui.QFileDialog.getOpenFileName(self,
+            gettext('Select editor executable'),
+            '/')
+        if filename:
+            self.editorEdit.setText(filename)
