@@ -53,13 +53,12 @@ def show_diff(arg_provider, ext_diff=None, parent_window=None):
         if parent_window:
             parent_window.windows.append(window)
     else:
-        args=["diff",
-              "--using=%s" % ext_diff]
+        args=["diff", "--using", ext_diff]  # NEVER USE --using=xxx, ALWAYS --using xxx
         # This should be move to after the window has been shown.
         dir, extra_args = arg_provider.get_ext_diff_args(
                                         QtCore.QCoreApplication.processEvents)
         args.extend(extra_args)
-        
+
         window = SimpleSubProcessDialog("External Diff",
                                         desc=ext_diff,
                                         args=args,
@@ -130,4 +129,3 @@ class DiffButtons(QtGui.QWidget):
         if ext_diff is None:
             ext_diff = QtCore.QString(default_diff)
         self.emit(QtCore.SIGNAL("triggered(QString)"), ext_diff)
-
