@@ -319,8 +319,11 @@ class SidebySideDiffView(QtGui.QSplitter):
                         if not p:
                             return []
                         lexer = get_lexer_for_filename(path, stripnl=False)
-                        tokens = list(split_tokens_at_lines(lex(d, lexer)))
-                        QtCore.QCoreApplication.processEvents() 
+                        tokens = []
+                        for token in split_tokens_at_lines(lex(d, lexer)):
+                            tokens.append(token)
+                            if len(token) % 100 == 0:
+                                QtCore.QCoreApplication.processEvents()
                         return tokens
 
                     display_lines = [getTokens(p, d, path)
@@ -451,6 +454,9 @@ class SidebySideDiffView(QtGui.QSplitter):
                     cursor = cursors[1]
                 for l in exlines:
                     insertLine(cursor, l)
+                
+                if i % 100 == 0:
+                    QtCore.QCoreApplication.processEvents()
         else:
             heights = [0,0]
             is_images = [False, False]
@@ -669,8 +675,6 @@ class SimpleDiffView(QtGui.QTextBrowser):
                 i0, i1, j0, j1 = group[0][1], group[-1][2], group[0][3], group[-1][4]
                 self.cursor.insertText("@@ -%d,%d +%d,%d @@\n" % (i0+1, i1-i0, j0+1, j1-j0), self.monospacedHunkFormat)
                 for tag, i0, i1, j0, j1 in group:
-                    ni = i1 - i0
-                    nj = j1 - j0
                     if tag == "equal":
                         text = "".join(" " + l for l in a[i0:i1])
                         self.cursor.insertText(text, self.monospacedFormat)
