@@ -58,6 +58,16 @@ from bzrlib.plugins.qbzr.lib.util import (
     )
 from bzrlib.plugins.qbzr.lib.uifactory import ui_current_widget
 from bzrlib.plugins.qbzr.lib.trace import reports_exception
+from bzrlib.plugins.qbzr.lib.encoding_selector import EncodingSelector
+
+try:
+    from bzrlib.errors import FileTimestampUnavailable
+except ImportError:
+    # FileTimestampUnavailable is available only in bzr 2.1.0rc1 and up
+    from bzrlib.errors import BzrError
+    class FileTimestampUnavailable(BzrError):
+        """Fake FileTimestampUnavailable error for older bzr."""
+        pass
 
 
 def get_file_lines_from_tree(tree, file_id):
@@ -154,6 +164,23 @@ class DiffWindow(QBzrWindow):
                      QtCore.SIGNAL("clicked(bool)"),
                      self.click_unidiff)
 
+        def on_left_encoding_changed(encoding):
+            if self.branches:
+                get_set_encoding(encoding, self.branches[0])
+            self.click_refresh()
+        self.encoding_selector_left = EncodingSelector(encoding,
+                                            gettext("Left side encoding:"),
+                                            on_left_encoding_changed)
+
+        def on_right_encoding_changed(encoding):
+            if self.branches:
+                get_set_encoding(encoding, self.branches[1])
+            self.click_refresh()
+        self.encoding_selector_right = EncodingSelector(encoding,
+                                            gettext("Right side encoding:"),
+                                            on_right_encoding_changed)
+
+
         complete = QtGui.QCheckBox (gettext("Complete"),
                                             self.centralwidget)
         self.connect(complete,
@@ -184,6 +211,8 @@ class DiffWindow(QBzrWindow):
         hbox.addWidget(complete)
         if has_ext_diff():
             hbox.addWidget(ext_diff_button)
+        hbox.addWidget(self.encoding_selector_left)
+        hbox.addWidget(self.encoding_selector_right)
         hbox.addWidget(buttonbox)
         vbox.addLayout(hbox)
 
@@ -208,7 +237,6 @@ class DiffWindow(QBzrWindow):
             self.throbber.hide()
 
     def load_branch_info(self):
-        
         (tree1, tree2,
          branch1, branch2,
          specific_files) = self.arg_provider.get_diff_window_args(self.processEvents)
@@ -219,8 +247,8 @@ class DiffWindow(QBzrWindow):
         
         self.set_diff_title()
         
-        self.encodings = (get_set_encoding(self.encoding, branch1),
-                          get_set_encoding(self.encoding, branch2))
+        self.encoding_selector_left.encoding = get_set_encoding(self.encoding, branch1)
+        self.encoding_selector_right.encoding = get_set_encoding(self.encoding, branch2)
         self.processEvents()
     
     def set_diff_title(self):
@@ -243,7 +271,7 @@ class DiffWindow(QBzrWindow):
                 title.append(self.filter_options.to_str())
 
         self.set_title_and_icon(title)
-        self.processEvents()        
+        self.processEvents()
 
     def load_diff(self):
         self.refresh_button.setEnabled(False)
@@ -302,6 +330,9 @@ class DiffWindow(QBzrWindow):
                                 # If we get ENOENT error then probably we trigger
                                 # bug #251532 in bzrlib. Take current time instead
                                 dates[ix] = time.time()
+                            except FileTimestampUnavailable:
+                                # ghosts around us (see Bug #513096)
+                                dates[ix] = 0  # using 1970/1/1 instead
 
                     properties_changed = [] 
                     if bool(executable[0]) != bool(executable[1]):
@@ -346,12 +377,15 @@ class DiffWindow(QBzrWindow):
                                     groups = list([matcher.get_opcodes()])
                                 else:
                                     groups = list(matcher.get_grouped_opcodes())
-                            try:
-                                lines = [[i.decode(encoding) for i in l]
-                                        for l, encoding in zip(lines, self.encodings)]
-                            except UnicodeDecodeError, e:
-                                trace.note('Failed to decode using %s, falling back to latin1' % e.encoding)
-                                lines = [[i.decode('latin1') for i in l] for l in lines]
+                            ulines = []
+                            for l, encoding in zip(lines, [self.encoding_selector_left.encoding,
+                                                           self.encoding_selector_right.encoding]):
+                                try:
+                                    ulines.append([i.decode(encoding) for i in l])
+                                except UnicodeDecodeError, e:
+                                    trace.note('Failed to decode using %s, falling back to latin1', e.encoding)
+                                    ulines.append([i.decode('latin1') for i in l])
+                            lines = ulines
                             data = ((),())
                         else:
                             groups = []
@@ -413,4 +447,5 @@ class DiffWindow(QBzrWindow):
         return False
     
     def ext_diff_triggered(self, ext_diff):
+        """@param ext_diff: path to external diff executable."""
         show_diff(self.arg_provider, ext_diff=ext_diff, parent_window = self)
