@@ -20,14 +20,16 @@
 from PyQt4 import QtCore, QtGui, Qt
 
 from bzrlib.bzrdir import BzrDir
+from bzrlib.revision import NULL_REVISION
+from bzrlib.revisionspec import RevisionSpec
 from bzrlib.plugins.qbzr.lib.revtreeview import (RevisionTreeView,
-                                                 RevNoItemDelegate)
+                                                 RevNoItemDelegate,
+                                                 get_text_color)
+from bzrlib.plugins.qbzr.lib.tag import TagWindow, CallBackTagWindow
 from bzrlib.plugins.qbzr.lib import logmodel
-from bzrlib.plugins.qbzr.lib.trace import *
-from bzrlib.plugins.qbzr.lib.util import (
-    runs_in_loading_queue,
-    )
 from bzrlib.plugins.qbzr.lib import diff
+from bzrlib.plugins.qbzr.lib.i18n import gettext
+
 
 class LogList(RevisionTreeView):
     """TreeView widget to show log with metadata and graph of revisions."""
@@ -48,8 +50,9 @@ class LogList(RevisionTreeView):
 
         self.setItemDelegateForColumn(logmodel.COL_MESSAGE,
                                       GraphTagsBugsItemDelegate(self))
+        self.rev_no_item_delegate = RevNoItemDelegate(parent=self)
         self.setItemDelegateForColumn(logmodel.COL_REV,
-                                      RevNoItemDelegate(parent=self))
+                                      self.rev_no_item_delegate)
         self.processEvents = processEvents
         self.throbber = throbber
 
@@ -72,8 +75,9 @@ class LogList(RevisionTreeView):
         header.setResizeMode(logmodel.COL_DATE, QtGui.QHeaderView.Interactive)
         header.setResizeMode(logmodel.COL_AUTHOR, QtGui.QHeaderView.Interactive)
         fm = self.fontMetrics()
-        # XXX Make this dynamic.
-        col_margin = 6
+        
+        col_margin = (self.style().pixelMetric(QtGui.QStyle.PM_FocusFrameHMargin,
+                                               None, self) + 1) *2
         header.resizeSection(logmodel.COL_REV,
                              fm.width("8888.8.888") + col_margin)
         header.resizeSection(logmodel.COL_DATE,
@@ -102,31 +106,31 @@ class LogList(RevisionTreeView):
                     diff_menu.setTitle(gettext("Show file &differences"))
                     self.context_menu.addMenu(diff_menu)
                     self.connect(diff_menu, QtCore.SIGNAL("triggered(QString)"),
-                                 self.show_diff_current_indexes)
+                                 self.show_diff_specified_files_ext)
                     
                     all_diff_menu = diff.ExtDiffMenu(self, set_default=False)
                     all_diff_menu.setTitle(gettext("Show all &differences"))
                     self.context_menu.addMenu(all_diff_menu)
                     self.connect(all_diff_menu, QtCore.SIGNAL("triggered(QString)"),
-                                 self.show_diff_current_indexes_all_files)
+                                 self.show_diff_ext)
                 else:
                     show_diff_action = self.context_menu.addAction(
                                         gettext("Show file &differences..."),
-                                        self.show_diff_current_indexes)
+                                        self.show_diff_specified_files)
                     self.context_menu.setDefaultAction(show_diff_action)
                     self.context_menu.addAction(
                                         gettext("Show all &differences..."),
-                                        self.show_diff_current_indexes_all_files)
+                                        self.show_diff)
             else:
                 if diff.has_ext_diff():
                     diff_menu = diff.ExtDiffMenu(self)
                     self.context_menu.addMenu(diff_menu)
                     self.connect(diff_menu, QtCore.SIGNAL("triggered(QString)"),
-                                 self.show_diff_current_indexes)
+                                 self.show_diff_ext)
                 else:
                     show_diff_action = self.context_menu.addAction(
                                         gettext("Show &differences..."),
-                                        self.show_diff_current_indexes)
+                                        self.show_diff)
                     self.context_menu.setDefaultAction(show_diff_action)
 
             self.connect(self,
@@ -134,7 +138,9 @@ class LogList(RevisionTreeView):
                          self.show_context_menu)
             
             self.context_menu.addAction(gettext("Show &tree..."),
-                                        self.show_revision_tree)        
+                                        self.show_revision_tree)
+            self.context_menu.addAction(gettext("Tag &revision..."),
+                                        self.tag_revision)
 
     def load_branch(self, branch, fileids, tree=None):
         self.throbber.show()
@@ -190,12 +196,32 @@ class LogList(RevisionTreeView):
         try:
             self.graph_provider.load_tags()
             self.log_model.load_graph_all_revisions()
+            self._adjust_revno_column()
         finally:
             self.graph_provider.unlock_branches()
         
         # Start later so that it does not run in the loading queue.
         QtCore.QTimer.singleShot(1, self.graph_provider.load_filter_file_id)
 
+    def _adjust_revno_column(self):
+        # update the data
+        max_mainline_digits = self.rev_no_item_delegate.set_max_revno(
+            self.graph_provider.max_mainline_revno)
+        # resize the column
+        header = self.header()
+        fm = self.fontMetrics()
+        col_margin = (self.style().pixelMetric(QtGui.QStyle.PM_FocusFrameHMargin,
+                                               None, self) + 1) *2
+        header.resizeSection(logmodel.COL_REV,
+            fm.width(("8"*max_mainline_digits)+".8.888") + col_margin)
+
+    def refresh_tags(self):
+        self.graph_provider.lock_read_branches()
+        try:
+            self.graph_provider.load_tags()
+        finally:
+            self.graph_provider.unlock_branches()
+
     def mousePressEvent (self, e):
         colapse_expand_click = False
         if e.button() & QtCore.Qt.LeftButton:
@@ -245,7 +271,6 @@ class LogList(RevisionTreeView):
         if not colapse_expand_click:
             QtGui.QTreeView.mouseMoveEvent(self, e)
 
-
     def keyPressEvent (self, e):
         e_key = e.key()
         if e_key in (QtCore.Qt.Key_Enter, QtCore.Qt.Key_Return) and self.view_commands:
@@ -286,28 +311,71 @@ class LogList(RevisionTreeView):
                                          (QtGui.QItemSelectionModel.Clear |
                                           QtGui.QItemSelectionModel.Select |
                                           QtGui.QItemSelectionModel.Rows))
+
+    def get_selection_indexes(self, index=None):
+        if index is None:
+            return sorted(self.selectionModel().selectedRows(0), 
+                          key=lambda x: x.row())
+        else:
+            return [index]
+    
+    def get_selection_and_merged_revids(self, index=None):
+        indexes = self.get_selection_indexes(index)
+        revids = set()
+        for index in indexes:
+            revid = str(index.data(logmodel.RevIdRole).toString())
+            revids.add(revid)
+            merges = [self.graph_provider.revisions[rev_index].revid
+                      for rev_index in
+                        self.graph_provider.revid_rev[revid].merges]
+            revids.update(set(merges))
+        return revids
+    
+    def get_selection_top_and_parent_revids_and_count(self, index=None):
+        indexes = self.get_selection_indexes(index)
+        if len(indexes) == 0:
+            return None, None
+        top_revid = str(indexes[0].data(logmodel.RevIdRole).toString())
+        bot_revid = str(indexes[-1].data(logmodel.RevIdRole).toString())
+        parents = self.graph_provider.graph_parents[bot_revid]
+        if parents:
+            # We need a ui to select which parent.
+            parent_revid = parents[0]
+            
+            # This is ugly. It is for the PendingMergesList in commit/revert.
+            if parent_revid == "root:":
+                parent_revid = self.graph_provider.graph.get_parent_map([bot_revid])[bot_revid][0]
+        else:
+            parent_revid = NULL_REVISION
+        return (top_revid, parent_revid), len(indexes)
     
     def set_search(self, str, field):
         self.graph_provider.set_search(str, field)
     
     def default_action(self, index=None):
-        if index is None:
-            self.show_diff_current_indexes()
-        else:
-            self.show_diff_index(index)
+        self.show_diff_specified_files()
+        
+    def tag_revision(self):
+        revid = str(self.currentIndex().data(logmodel.RevIdRole).toString())
+        revno = self.graph_provider.revid_rev[revid].revno_str
+        revs = [RevisionSpec.from_string(revno)]
+        branch = self.graph_provider.get_revid_branch(revid)
+        action = TagWindow.action_from_options(force=False, delete=False)
+        window = CallBackTagWindow(branch, self.refresh_tags, action=action, revision=revs)
+        window.show()
+        self.window().windows.append(window)
     
-    def show_diff(self, new_rev, old_rev,
+    def show_diff(self, index=None,
                   specific_files=None, specific_file_ids=None,
                   ext_diff=None):
-        new_revid = new_rev.revision_id
-        new_branch = self.graph_provider.get_revid_branch(new_revid)
         
-        if not old_rev.parent_ids:
-            old_revid = None
-            old_branch = new_branch
-        else:
-            old_revid = old_rev.parent_ids[0]
-            old_branch =  self.graph_provider.get_revid_branch(old_revid)
+        (new_revid, old_revid), count = \
+            self.get_selection_top_and_parent_revids_and_count(index)
+        if new_revid is None and old_revid is None:
+            # No revision selection.
+            return
+        new_branch = self.graph_provider.get_revid_branch(new_revid)
+        old_branch =  self.graph_provider.get_revid_branch(old_revid)
         
         arg_provider = diff.InternalDiffArgProvider(
                                         old_revid, new_revid,
@@ -317,43 +385,19 @@ class LogList(RevisionTreeView):
         
         diff.show_diff(arg_provider, ext_diff = ext_diff,
                        parent_window = self.window())
-
-    def show_diff_index(self, index):
-        """Show differences of a single revision from a index."""
-        revid = str(index.data(logmodel.RevIdRole).toString())
-        rev = self.graph_provider.load_revisions([revid])[revid]
-        if self.graph_provider.fileids:
-            self.show_diff(rev, rev,
-                           specific_file_ids=self.graph_provider.fileids)
-        else:
-            self.show_diff(rev, rev)
     
-    def show_diff_current_indexes(self, ext_diff=None,
-                                  only_specified_files=True):
-        """Show differences of the selected range or of a single revision"""
-        # Find 1 index for each row.
-        rows = {}
-        for index in self.selectedIndexes():
-            if index.row() not in rows:
-                rows[index.row()] = index
-        indexes = rows.values()
-        if not indexes:
-            # the list is empty
-            return
-        revid1 = str(indexes[0].data(logmodel.RevIdRole).toString())
-        revid2 = str(indexes[-1].data(logmodel.RevIdRole).toString())
-        revs = self.graph_provider.load_revisions([revid1, revid2])
-        rev1 = revs[revid1]
-        rev2 = revs[revid2]
-        if only_specified_files and self.graph_provider.fileids:
-            self.show_diff(rev1, rev2, ext_diff=ext_diff,
+    def show_diff_specified_files(self, ext_diff=None):
+        if self.graph_provider.fileids:
+            self.show_diff(ext_diff=ext_diff,
                            specific_file_ids = self.graph_provider.fileids)
         else:
-            self.show_diff(rev1, rev2, ext_diff=ext_diff)
+            self.show_diff(ext_diff=ext_diff)
     
-    def show_diff_current_indexes_all_files(self, ext_diff=None):
-        self.show_diff_current_indexes(ext_diff=ext_diff,
-                                       only_specified_files=False)
+    def show_diff_ext(self, ext_diff):
+        self.show_diff(ext_diff=ext_diff)
+
+    def show_diff_specified_files_ext(self, ext_diff=None):
+        self.show_diff_specified_files(ext_diff=ext_diff)
     
     def show_revision_tree(self):
         from bzrlib.plugins.qbzr.lib.browse import BrowseWindow
@@ -368,8 +412,8 @@ class LogList(RevisionTreeView):
     def show_context_menu(self, pos):
         self.context_menu.popup(self.viewport().mapToGlobal(pos))
 
-    
-class GraphTagsBugsItemDelegate(QtGui.QItemDelegate):
+
+class GraphTagsBugsItemDelegate(QtGui.QStyledItemDelegate):
 
     _tagColor = QtGui.QColor(80, 128, 32)
     _bugColor = QtGui.QColor(164, 0, 0)
@@ -381,7 +425,7 @@ class GraphTagsBugsItemDelegate(QtGui.QItemDelegate):
     def paint(self, painter, option, index):
         node = index.data(logmodel.GraphNodeRole)
         if node.isValid():
-            self.drawGraph = True
+            draw_graph = True
             self.node = node.toList()
             self.lines = index.data(logmodel.GraphLinesRole).toList()
             self.twisty_state = index.data(logmodel.GraphTwistyStateRole)
@@ -392,7 +436,7 @@ class GraphTagsBugsItemDelegate(QtGui.QItemDelegate):
             else:
                 self.prevLines = []
         else:
-            self.drawGraph = False
+            draw_graph = False
         
         self.labels = []
         # collect branch tags
@@ -407,59 +451,27 @@ class GraphTagsBugsItemDelegate(QtGui.QItemDelegate):
         for bug in index.data(logmodel.BugIdsRole).toStringList():
             self.labels.append(
                 (bug, self._bugColor))
-        QtGui.QItemDelegate.paint(self, painter, option, index)
-    
-    def get_color(self, color, back):
-        qcolor = QtGui.QColor()
-        if color == 0:
-            if back:
-                qcolor.setHsvF(0,0,0.8)
-            else:
-                qcolor.setHsvF(0,0,0)
-        else:
-            h = float(color % 6) / 6
-            if back:
-                qcolor.setHsvF(h,0.4,1)
-            else:
-                qcolor.setHsvF(h,1,0.7)
         
-        return qcolor
-    
-    def drawLine(self, painter, pen, rect, boxsize, mid, height,
-                 start, end, color, direct):
-        pen.setColor(self.get_color(color,False))
-        if direct:
-            pen.setStyle(QtCore.Qt.SolidLine)
-        else:
-            pen.setStyle(QtCore.Qt.DotLine)            
-        painter.setPen(pen)
-        startx = rect.x() + boxsize * start + boxsize / 2 
-        endx = rect.x() + boxsize * end + boxsize / 2 
+        option = QtGui.QStyleOptionViewItemV4(option)
+        self.initStyleOption(option, index)
+        widget = self.parent()
+        style = widget.style()
         
-        path = QtGui.QPainterPath()
-        path.moveTo(QtCore.QPointF(startx, mid - height / 2))
+        text_margin = style.pixelMetric(QtGui.QStyle.PM_FocusFrameHMargin,
+                                        None, widget) + 1
         
-        if start - end == 0 :
-            path.lineTo(QtCore.QPointF(endx, mid + height / 2)) 
-        else:
-            path.cubicTo(QtCore.QPointF(startx, mid - height / 5),
-                         QtCore.QPointF(startx, mid - height / 5),
-                         QtCore.QPointF(startx + (endx - startx) / 2, mid))
-
-            path.cubicTo(QtCore.QPointF(endx, mid + height / 5),
-                         QtCore.QPointF(endx, mid + height / 5),
-                         QtCore.QPointF(endx, mid + height / 2 + 1))
-        painter.drawPath(path)
-        pen.setStyle(QtCore.Qt.SolidLine)
-
-    def drawDisplay(self, painter, option, rect, text):
-
         if not hasattr(self, '_usingGtkStyle'):
-            self._usingGtkStyle = Qt.qApp.style().objectName() == 'gtk+'
-            self._usingQt45 = Qt.qVersion() >= '4.5' 
-
+            self._usingGtkStyle = style.objectName() == 'gtk+'
+            self._usingQt45 = Qt.qVersion() >= '4.5'
+        
+        painter.save()
+        painter.setClipRect(option.rect)
+        style.drawPrimitive(QtGui.QStyle.PE_PanelItemViewItem,
+                            option, painter, widget)
+        
         graphCols = 0
-        if self.drawGraph:
+        rect = option.rect
+        if draw_graph:
             painter.save()
             try:
                 painter.setRenderHint(QtGui.QPainter.Antialiasing)            
@@ -539,14 +551,14 @@ class GraphTagsBugsItemDelegate(QtGui.QItemDelegate):
                 
             finally:
                 painter.restore()
-            rect.adjust( (graphCols + 1.7) * boxsize, 0, 0, 0)
-
+            rect.adjust( (graphCols + 1.5) * boxsize, 0, 0, 0)        
         painter.save()
+        
+        x = 0
         try:
             tagFont = QtGui.QFont(option.font)
             tagFont.setPointSizeF(tagFont.pointSizeF() * 9 / 10)
     
-            x = 0
             for label, color in self.labels:
                 tagRect = rect.adjusted(1, 1, -1, -1)
                 tagRect.setWidth(QtGui.QFontMetrics(tagFont).width(label) + 6)
@@ -565,10 +577,65 @@ class GraphTagsBugsItemDelegate(QtGui.QItemDelegate):
                     painter.drawText(tagRect.left() + 3, tagRect.bottom() - option.fontMetrics.descent(), label)
                 else:
                     painter.drawText(tagRect.left() + 3, tagRect.bottom() - option.fontMetrics.descent() + 1, label)
-                x += tagRect.width() + 3
+                x += tagRect.width() + text_margin
         finally:
             painter.restore()
+        rect.adjust(x, 0, 0, 0)
         
-        rect.adjust( x, 0, 0, 0)
+        if not option.text.isEmpty():
+            painter.setPen(get_text_color(option, style))
+            text_rect = rect.adjusted(0, 0, -text_margin, 0)
+            painter.setFont(option.font)
+            fm = painter.fontMetrics()
+            text_width = fm.width(option.text)
+            text = option.text
+            if text_width > text_rect.width():
+                text = self.elidedText(fm, text_rect.width(),
+                                       QtCore.Qt.ElideRight, text)
+            
+            painter.drawText(text_rect, QtCore.Qt.AlignLeft, text)
         
-        return QtGui.QItemDelegate.drawDisplay(self, painter, option, rect, text)
+        painter.restore()
+    
+    def get_color(self, color, back):
+        qcolor = QtGui.QColor()
+        if color == 0:
+            if back:
+                qcolor.setHsvF(0,0,0.8)
+            else:
+                qcolor.setHsvF(0,0,0)
+        else:
+            h = float(color % 6) / 6
+            if back:
+                qcolor.setHsvF(h,0.4,1)
+            else:
+                qcolor.setHsvF(h,1,0.7)
+        
+        return qcolor
+    
+    def drawLine(self, painter, pen, rect, boxsize, mid, height,
+                 start, end, color, direct):
+        pen.setColor(self.get_color(color,False))
+        if direct:
+            pen.setStyle(QtCore.Qt.SolidLine)
+        else:
+            pen.setStyle(QtCore.Qt.DotLine)            
+        painter.setPen(pen)
+        startx = rect.x() + boxsize * start + boxsize / 2 
+        endx = rect.x() + boxsize * end + boxsize / 2 
+        
+        path = QtGui.QPainterPath()
+        path.moveTo(QtCore.QPointF(startx, mid - height / 2))
+        
+        if start - end == 0 :
+            path.lineTo(QtCore.QPointF(endx, mid + height / 2)) 
+        else:
+            path.cubicTo(QtCore.QPointF(startx, mid - height / 5),
+                         QtCore.QPointF(startx, mid - height / 5),
+                         QtCore.QPointF(startx + (endx - startx) / 2, mid))
+
+            path.cubicTo(QtCore.QPointF(endx, mid + height / 5),
+                         QtCore.QPointF(endx, mid + height / 5),
+                         QtCore.QPointF(endx, mid + height / 2 + 1))
+        painter.drawPath(path)
+        pen.setStyle(QtCore.Qt.SolidLine)
