@@ -21,20 +21,22 @@
 
 import codecs
 import os
+import re
+import shlex
 import signal
 import sys
 import tempfile
+import thread
 
 from PyQt4 import QtCore, QtGui
 
-from bzrlib import osutils, progress, errors
-
-try:
-    # this works with bzr 1.16+
-    from bzrlib import bencode
-except ImportError:
-    # this works with bzr 1.15-
-    from bzrlib.util import bencode
+from bzrlib import (
+    bencode,
+    commands,
+    osutils,
+    ui,
+    )
+from bzrlib.option import Option
 
 from bzrlib.plugins.qbzr.lib import MS_WINDOWS
 from bzrlib.plugins.qbzr.lib.i18n import gettext
@@ -47,7 +49,6 @@ from bzrlib.plugins.qbzr.lib.util import (
     StandardButton,
     ensure_unicode,
     )
-from bzrlib.plugins.qbzr.lib.uifactory import ui_current_widget
 from bzrlib.plugins.qbzr.lib.trace import (
    report_exception,
    SUB_LOAD_METHOD)
@@ -55,6 +56,13 @@ from bzrlib.plugins.qbzr.lib.trace import (
 from bzrlib.ui.text import TextProgressView, TextUIFactory
 
 
+# Subprocess service messages markers
+SUB_PROGRESS = "qbzr:PROGRESS:"
+SUB_GETPASS = "qbzr:GETPASS:"
+SUB_GETUSER = "qbzr:GETUSER:"
+SUB_GETBOOL = "qbzr:GETBOOL:"
+
+
 class SubProcessWindowBase(object):
 
     def __init_internal__(self, title,
@@ -117,13 +125,18 @@ class SubProcessWindowBase(object):
                                QtCore.SIGNAL("subprocessFinished(bool)"),
                                cancelButton,
                                QtCore.SLOT("setDisabled(bool)"))
-        
+
         # ok button gets enabled when we fail.
         QtCore.QObject.connect(self,
                                QtCore.SIGNAL("subprocessFailed(bool)"),
                                okButton,
                                QtCore.SLOT("setDisabled(bool)"))
 
+        # Change the ok button to 'retry' if we fail.
+        QtCore.QObject.connect(self,
+                               QtCore.SIGNAL("subprocessFailed(bool)"),
+                               lambda failed: okButton.setText(gettext('&Retry')))
+
         self.buttonbox = QtGui.QDialogButtonBox(self)
         self.buttonbox.addButton(okButton,
             QtGui.QDialogButtonBox.AcceptRole)
@@ -141,8 +154,13 @@ class SubProcessWindowBase(object):
         status_layout.setContentsMargins(0, 0, 0, 0)
         status_layout.addWidget(self.process_widget)
         return status_group_box
-        
+
     def make_default_layout_widgets(self):
+        """Yields widgets to add to main dialog layout: status and button boxes.
+        Status box has progress bar and console area.
+        Button box has 2 buttons: OK and Cancel (after successfull command 
+        execution there will be Close and Cancel).
+        """
         yield self.make_default_status_box()
         yield self.buttonbox
 
@@ -171,17 +189,17 @@ class SubProcessWindowBase(object):
             except:
                 report_exception(type=SUB_LOAD_METHOD, window=self.window())
                 return
-            
+
             self.emit(QtCore.SIGNAL("subprocessStarted(bool)"), True)
             self.emit(QtCore.SIGNAL("disableUi(bool)"), True)
             self.do_start()
-    
+
     def do_start(self):
         if self._check_args():
             self.process_widget.do_start(self.dir, *self.args)
         else:
             self.on_failed()
-    
+
     def do_reject(self):
         if self.process_widget.is_running():
             self.process_widget.abort()
@@ -191,7 +209,7 @@ class SubProcessWindowBase(object):
     def on_finished(self):
         if hasattr(self, 'setResult'):
             self.setResult(QtGui.QDialog.Accepted)
-        
+
         self.emit(QtCore.SIGNAL("subprocessFinished(bool)"), True)
         self.emit(QtCore.SIGNAL("disableUi(bool)"), False)
 
@@ -279,7 +297,6 @@ class SubProcessDialog(SubProcessWindowBase, QBzrDialog):
             event.ignore()
 
 
-
 class SimpleSubProcessDialog(SubProcessDialog):
     """A concrete helper class of SubProcessDialog, which has a single label
     widget for displaying a simple description before executing a subprocess.
@@ -318,10 +335,10 @@ class SimpleSubProcessDialog(SubProcessDialog):
         # and add the subprocess widgets.
         for w in self.make_default_layout_widgets():
             layout.addWidget(w)
-        
+
         self.auto_start_show_on_failed = auto_start_show_on_failed
         QtCore.QTimer.singleShot(1, self.auto_start)
-    
+
     def auto_start(self):
         if self.auto_start_show_on_failed:
             QtCore.QObject.connect(self,
@@ -344,7 +361,7 @@ class SubProcessWidget(QtGui.QWidget):
         layout = QtGui.QVBoxLayout(self)
 
         message_layout = QtGui.QHBoxLayout()
-        
+
         self.progressMessage = QtGui.QLabel(self)
         #self.progressMessage.setWordWrap(True) -- this breaks minimal window size hint
         self.progressMessage.setText(gettext("Ready"))
@@ -352,7 +369,7 @@ class SubProcessWidget(QtGui.QWidget):
 
         self.transportActivity = QtGui.QLabel(self)
         message_layout.addWidget(self.transportActivity)
-        
+
         layout.addLayout(message_layout)
 
         self.progressBar = QtGui.QProgressBar(self)
@@ -380,15 +397,17 @@ class SubProcessWidget(QtGui.QWidget):
         self.connect(self.process,
             QtCore.SIGNAL("finished(int, QProcess::ExitStatus)"),
             self.onFinished)
-        
+
         self.defaultWorkingDir = self.process.workingDirectory ()
-        
+
         self.finished = False
         self.aborting = False
-        
+
         self.messageFormat = QtGui.QTextCharFormat()
         self.errorFormat = QtGui.QTextCharFormat()
         self.errorFormat.setForeground(QtGui.QColor('red'))
+        self.cmdlineFormat = QtGui.QTextCharFormat()
+        self.cmdlineFormat.setForeground(QtGui.QColor('blue'))
 
         if hide_progress:
             self.hide_progress()
@@ -398,57 +417,85 @@ class SubProcessWidget(QtGui.QWidget):
     def hide_progress(self):
         self.progressMessage.setHidden(True)
         self.progressBar.setHidden(True)
-    
+
     def is_running(self):
         return self.process.state() == QtCore.QProcess.Running or\
                self.process.state() == QtCore.QProcess.Starting
-    
+
     def do_start(self, workdir, *args):
         """Launch one bzr command.
         @param  workdir:    working directory for command.
                             Could be None to use self.defaultWorkingDir
         @param  args:   bzr command and its arguments
+        @type   args:   all arguments should be unicode strings (or ascii-only).
         """
         QtGui.QApplication.processEvents() # make sure ui has caught up
         self.start_multi(((workdir, args),))
-    
+
     def start_multi(self, commands):
         self.setProgress(0, [gettext("Starting...")], "")
         self.console.setFocus(QtCore.Qt.OtherFocusReason)
         self.commands = list(commands)
         self._start_next()
-    
+
     def _start_next(self):
+        """Start first command from self.commands queue."""
+        self._setup_stdout_stderr()
         self._delete_args_file()
         dir, args = self.commands.pop(0)
-        
-        def format_arg(a):
-            # Some code passes us QStrings. Convert to python string.
-            a = unicode(a)
-            # Don't quote revision arg.
-            if a.startswith("-r"):
-                return a
-            return '"%s"' % a.replace('"', '\\"')
-        
-        args = ' '.join(format_arg(a) for a in args)
-        if MS_WINDOWS:
-            # win32 has command-line length limit about 32K
-            if len(args) > 31000:
-                # save the args to the file
-                fname = self._create_args_file(args)
-                args = "@" + fname.replace('\\', '/')
+
+        # Log the command we about to execute
+        def format_args_for_log(args):
+            r = ['bzr']
+            for a in args:
+                a = unicode(a).translate({
+                        ord(u'\n'): u'\\n',
+                        ord(u'\r'): u'\\r',
+                        ord(u'\t'): u'\\t',
+                        })
+                if " " in a:
+                    r.append('"%s"' % a)
+                else:
+                    r.append(a)
+            s = ' '.join(r)
+            if len(s) > 128:  # XXX make it configurable?
+                s = s[:128] + '...'
+            return s
+        self.logMessageEx("Run command: "+format_args_for_log(args), "cmdline", self.stderr)
+
+        args = bencode_unicode(args)
+
+        # win32 has command-line length limit about 32K, but it seems 
+        # problems with command-line buffer limit occurs not only on windows.
+        # see bug https://bugs.launchpad.net/qbzr/+bug/396165
+        # on Linux I believe command-line is in utf-8,
+        # so we need to have some extra space
+        # when converting unicode -> utf8
+        if (len(args) > 10000       # XXX make the threshold configurable in qbzr.conf?
+            or re.search(r"(?:"
+                r"\n|\r"            # workaround for bug #517420
+                r"|\\\\"            # workaround for bug #528944
+                r")", args) is not None):
+            # save the args to the file
+            fname = self._create_args_file(args)
+            args = "@" + fname.replace('\\', '/')
 
         if dir is None:
             dir = self.defaultWorkingDir
 
-        self.process.setWorkingDirectory (dir)
-        self._setup_stdout_stderr()
+        self.process.setWorkingDirectory(dir)
         if getattr(sys, "frozen", None) is not None:
+            bzr_exe = sys.argv[0]
+            if sys.frozen == 'windows_exe':
+                # bzrw.exe
+                exe = os.path.join(os.path.dirname(sys.argv[0]), 'bzr.exe')
+                if os.path.isfile(exe):
+                    bzr_exe = exe
             self.process.start(
-                sys.argv[0], ['qsubprocess', args])
+                bzr_exe, ['qsubprocess', '--bencode', args])
         else:
             self.process.start(
-                sys.executable, [sys.argv[0], 'qsubprocess', args])
+                sys.executable, [sys.argv[0], 'qsubprocess', '--bencode', args])
 
     def _setup_stdout_stderr(self):
         if self.stdout is None:
@@ -470,10 +517,10 @@ class SubProcessWidget(QtGui.QWidget):
                 self.setProgress(None, [gettext("Aborting...")])
             else:
                 self.process.terminate()
-    
+
     def abort_futher_processes(self):
         self.commands = []
-    
+
     def setProgress(self, progress, messages, transport_activity=None):
         if progress is not None:
             self.progressBar.setValue(progress)
@@ -487,54 +534,94 @@ class SubProcessWidget(QtGui.QWidget):
         self.progressMessage.setText(text)
         if transport_activity is not None:
             self.transportActivity.setText(transport_activity)
-    
+
     def readStdout(self):
         # ensure we read from subprocess plain string
         data = str(self.process.readAllStandardOutput())
         # we need unicode for all strings except bencoded streams
         for line in data.splitlines():
-            if line.startswith("qbzr:PROGRESS:"):
-                # but we have to ensure we have unicode after bdecode
-                progress, transport_activity, messages = map(ensure_unicode, bencode.bdecode(line[14:]))
-                self.setProgress(progress, messages, transport_activity)
-            elif line.startswith("qbzr:GETPASS:"):
-                prompt = bencode.bdecode(line[13:]).decode('utf-8')
+            if line.startswith(SUB_PROGRESS):
+                try:
+                    # map(ensure_unicode, ...) to ensure that we have unicode after bdecode
+                    progress, transport_activity, messages = map(ensure_unicode,
+                        bencode.bdecode(line[len(SUB_PROGRESS):]))
+                except ValueError, e:
+                    # we got malformed data from qsubprocess (bencode failed to decode)
+                    # so just show it in the status console
+                    self.logMessageEx("qsubprocess error: "+str(e), "error", self.stderr)
+                    self.logMessageEx(line.decode(self.encoding), "error", self.stderr)
+                else:
+                    self.setProgress(progress, messages, transport_activity)
+            elif line.startswith(SUB_GETPASS):
+                prompt = bdecode_prompt(line[len(SUB_GETPASS):])
                 passwd, ok = QtGui.QInputDialog.getText(self,
                                                         gettext("Enter Password"),
                                                         prompt,
                                                         QtGui.QLineEdit.Password)
                 data = unicode(passwd).encode('utf-8'), int(ok)
-                self.process.write("qbzr:GETPASS:"+bencode.bencode(data)+"\n")
+                self.process.write(SUB_GETPASS + bencode.bencode(data) + "\n")
                 if not ok:
                     self.abort_futher_processes()
+            elif line.startswith(SUB_GETUSER):
+                prompt = bdecode_prompt(line[len(SUB_GETUSER):])
+                passwd, ok = QtGui.QInputDialog.getText(self,
+                                                        gettext("Enter Username"),
+                                                        prompt)
+                data = unicode(passwd).encode('utf-8'), int(ok)
+                self.process.write(SUB_GETUSER + bencode.bencode(data) + "\n")
+                if not ok:
+                    self.abort_futher_processes()
+            elif line.startswith(SUB_GETBOOL):
+                prompt = bdecode_prompt(line[len(SUB_GETBOOL):])
+                button = QtGui.QMessageBox.question(
+                    self, "Bazaar", prompt,
+                    QtGui.QMessageBox.Yes, QtGui.QMessageBox.No)
+                
+                data = (button == QtGui.QMessageBox.Yes)
+                self.process.write(SUB_GETBOOL + bencode.bencode(data) + "\n")
             else:
-                line = line.decode(self.encoding)
-                self.logMessage(line)
-                if not self.ui_mode:
-                    self.stdout.write(line)
-                    self.stdout.write("\n")
-    
+                line = line.decode(self.encoding, 'replace')
+                self.logMessageEx(line, 'plain', self.stdout)
+
     def readStderr(self):
         data = str(self.process.readAllStandardError()).decode(self.encoding, 'replace')
         if data:
             self.emit(QtCore.SIGNAL("error()"))
-        
+
         for line in data.splitlines():
             error = line.startswith("bzr: ERROR:")
-            self.logMessage(line, error)
-            if not self.ui_mode:
-                self.stderr.write(line)
-                self.stderr.write("\n")
+            self.logMessage(line, error, self.stderr)
 
-    def logMessage(self, message, error=False):
+    def logMessage(self, message, error=False, terminal_stream=None):
+        kind = 'plain'
         if error:
+            kind = 'error'
+        self.logMessageEx(message, kind, terminal_stream)
+
+    def logMessageEx(self, message, kind="plain", terminal_stream=None):
+        """Write message to console area.
+        @param kind: kind of message used for selecting style of formatting.
+            Possible kind values:
+                * plain = usual message, written in default style;
+                * error = error message, written in red;
+                * cmdline = show actual command-line, written in blue.
+        @param terminal_stream: if we working in non --ui-mode
+            the message can be echoed to real terminal via specified
+            terminal_stream (e.g. sys.stdout or sys.stderr)
+        """
+        if kind == 'error':
             format = self.errorFormat
+        elif kind == 'cmdline':
+            format = self.cmdlineFormat
         else:
             format = self.messageFormat
-        self.console.setCurrentCharFormat(format);
-        self.console.append(message);
+        self.console.setCurrentCharFormat(format)
+        self.console.append(message)
         scrollbar = self.console.verticalScrollBar()
         scrollbar.setValue(scrollbar.maximum())
+        if not self.ui_mode and terminal_stream:
+            terminal_stream.write(message)
+            terminal_stream.write('\n')
 
     def reportProcessError(self, error):
         self.aborting = False
@@ -585,14 +672,14 @@ class SubProcessWidget(QtGui.QWidget):
         if self._args_file:
             try:
                 os.unlink(self._args_file)
-            except (IOError, OSError), e:
+            except (IOError, OSError):
                 pass
             else:
                 self._args_file = None
 
 
 class SubprocessProgressView (TextProgressView):
-    
+
     def __init__(self, term_file):
         TextProgressView.__init__(self, term_file)
         # The TextProgressView does not show the transport activity untill
@@ -600,7 +687,7 @@ class SubprocessProgressView (TextProgressView):
         # transport activity before a progress update would cause artifactes to
         # remain on the screen. We don't have to worry about that
         self._have_output = True
-    
+
     def _repaint(self):
         if self._last_task:
             task_msg = self._format_task(self._last_task)
@@ -612,30 +699,25 @@ class SubprocessProgressView (TextProgressView):
         else:
             task_msg = ''
             progress = 0
-        
+
         trans = self._last_transport_msg
-        
-        self._term_file.write('qbzr:PROGRESS:' + bencode.bencode((progress,
-                              trans, task_msg)) + '\n')
+
+        self._term_file.write(
+            SUB_PROGRESS + bencode.bencode((progress, trans, task_msg)) + '\n')
         self._term_file.flush()
-    
+
     def clear(self):
         pass
 
 
 class SubprocessUIFactory(TextUIFactory):
 
-    def __init__(self, stdin=None, stdout=None, stderr=None):
-        
-        TextUIFactory.__init__(self, stdin=stdin, stdout=stdout, stderr=stderr)
-        
-        # This is to be compatabile with bzr < rev 4558
-        if not hasattr(TextUIFactory, "make_progress_view"):
-            self._progress_view = SubprocessProgressView(self.stdout)
-    
     def make_progress_view(self):
         return SubprocessProgressView(self.stdout)
-
+    
+    # This is to be compatabile with bzr < rev 4558
+    _make_progress_view = make_progress_view
+    
     def clear_term(self):
         """Prepare the terminal for output.
 
@@ -643,18 +725,72 @@ class SubprocessUIFactory(TextUIFactory):
         cursor at the leftmost position."""
         pass
 
-    def get_password(self, prompt='', **kwargs):
-        prompt = prompt % kwargs
-        self.stdout.write('qbzr:GETPASS:' + bencode.bencode(prompt.encode('utf-8')) + '\n')
+    def _get_answer_from_main(self, name, arg):
+        self.stdout.write(name + bencode_prompt(arg) + '\n')
         self.stdout.flush()
         line = self.stdin.readline()
-        if line.startswith('qbzr:GETPASS:'):
-            passwd, accepted = bencode.bdecode(line[13:].rstrip('\r\n'))
-            if accepted:
-                return passwd
-            else:
-                raise KeyboardInterrupt()
-        raise Exception("Did not recive a password from the main process.")
+        if line.startswith(name):
+            return bencode.bdecode(line[len(name):].rstrip('\r\n'))
+        raise Exception("Did not recive a answer from the main process.")
+    
+    def get_password(self, prompt='', **kwargs):
+        prompt = prompt % kwargs
+        passwd, accepted = self._get_answer_from_main(SUB_GETPASS, prompt)
+        if accepted:
+            return passwd
+        else:
+            raise KeyboardInterrupt()
+    
+    def get_username(self, prompt='', **kwargs):
+        prompt = prompt % kwargs
+        username, accepted = self._get_answer_from_main(SUB_GETUSER, prompt)
+        if accepted:
+            return username
+        else:
+            raise KeyboardInterrupt()
+
+    def get_boolean(self, prompt):
+        return self._get_answer_from_main(SUB_GETBOOL, prompt+'?')
+
+
+# [bialix 2010/02/04] body of cmd_qsubprocess has moved from commands.py
+# to see annotation of cmd_qsubprocess before move use:
+#     bzr qannotate commands.py -r1117
+
+def run_subprocess_command(cmd, bencoded=False):
+    """The actual body of qsubprocess.
+    Running specified bzr command in the subprocess.
+    @param cmd: string with command line to run.
+    @param bencoded: either cmd_str is bencoded list or not.
+
+    NOTE: if cmd starts with @ sign then it used as name of the file
+    where actual command line string is saved (utf-8 encoded).
+    """
+    if MS_WINDOWS:
+        thread.start_new_thread(windows_emulate_ctrl_c, ())
+    else:
+        signal.signal(signal.SIGINT, sigabrt_handler)
+    ui.ui_factory = SubprocessUIFactory(stdin=sys.stdin,
+                                        stdout=sys.stdout,
+                                        stderr=sys.stderr)
+    if cmd.startswith('@'):
+        fname = cmd[1:]
+        f = open(fname, 'rb')
+        try:
+            cmd_utf8 = f.read()
+        finally:
+            f.close()
+    else:
+        cmd_utf8 = cmd.encode('utf8')
+    if not bencoded:
+        argv = [unicode(p, 'utf-8') for p in shlex.split(cmd_utf8)]
+    else:
+        argv = [unicode(p, 'utf-8') for p in bencode.bdecode(cmd_utf8)]
+    return commands.run_bzr(argv)
+
+
+def sigabrt_handler(signum, frame):
+    raise KeyboardInterrupt()
 
 
 if MS_WINDOWS:
@@ -671,9 +807,9 @@ if MS_WINDOWS:
                     ("hThread", HANDLE),
                     ("dwProcessID", DWORD),
                     ("dwThreadID", DWORD)]
-    
+
     LPPROCESS_INFORMATION = POINTER(PROCESS_INFORMATION)
-    
+
     def get_child_pid(voidptr):
         lp = cast(int(voidptr), LPPROCESS_INFORMATION)
         return lp.contents.dwProcessID
@@ -689,3 +825,29 @@ if MS_WINDOWS:
         finally:
             ev.Close()
 
+    def windows_emulate_ctrl_c():
+        """To emulate Ctrl+C on Windows we have to wait some global event,
+        and once it will trigger we will try to interrupt main thread.
+        IMPORTANT: This function should be invoked as separate thread!
+        """
+        import win32event
+        ev = win32event.CreateEvent(None, 0, 0, get_event_name(os.getpid()))
+        try:
+            win32event.WaitForSingleObject(ev, win32event.INFINITE)
+        finally:
+            ev.Close()
+        thread.interrupt_main()
+
+
+def bencode_unicode(args):
+    """Bencode list of unicode strings as list of utf-8 strings and converting
+    resulting string to unicode.
+    """
+    args_utf8 = bencode.bencode([unicode(a).encode('utf-8') for a in args])
+    return unicode(args_utf8, 'utf-8')
+
+def bencode_prompt(arg):
+    return bencode.bencode(arg.encode('unicode-escape'))
+
+def bdecode_prompt(s):
+    return bencode.bdecode(s).decode('unicode-escape')
