@@ -19,13 +19,13 @@
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
 import os
-import re
+import shlex
 import sys
 import itertools
 
 from PyQt4 import QtCore, QtGui
 
-from bzrlib.revision import Revision, CURRENT_REVISION
+from bzrlib.revision import Revision
 from bzrlib.config import (
     GlobalConfig,
     IniBasedConfig,
@@ -39,34 +39,15 @@ from bzrlib import (
     )
 from bzrlib.util.configobj import configobj
 
+from bzrlib.plugins.qbzr.lib import MS_WINDOWS
 from bzrlib.plugins.qbzr.lib import trace
-from bzrlib.plugins.qbzr.lib import i18n
-from bzrlib.plugins.qbzr.lib.i18n import gettext, N_, ngettext
+from bzrlib.plugins.qbzr.lib.i18n import gettext, N_
 
 # pyflakes says this is not needed, but it is.
 import bzrlib.plugins.qbzr.lib.resources
 
-from bzrlib.lazy_import import lazy_import
-lazy_import(globals(), '''
 from bzrlib import errors
 from bzrlib.workingtree import WorkingTree
-''')
-
-
-_email_re = lazy_regex.lazy_compile(r'([a-z0-9_\-.+]+@[a-z0-9_\-.+]+)', re.IGNORECASE)
-_link1_re = lazy_regex.lazy_compile(r'([\s>])(https?)://([^\s<>{}()]+[^\s.,<>{}()])', re.IGNORECASE)
-_link2_re = lazy_regex.lazy_compile(r'(\s)www\.([a-z0-9\-]+)\.([a-z0-9\-.\~]+)((?:/[^ <>{}()\n\r]*[^., <>{}()\n\r]?)?)', re.IGNORECASE)
-_tag_re = lazy_regex.lazy_compile(r'[, ]')
-
-
-def htmlize(text):
-    text = htmlencode(text)
-    text = text.replace("\n", '<br />')
-    text = _email_re.sub('<a href="mailto:\\1">\\1</a>', text)
-    text = _link1_re.sub('\\1<a href="\\2://\\3">\\2://\\3</a>', text)
-    text = _link2_re.sub('\\1<a href="http://www.\\2.\\3\\4">www.\\2.\\3\\4</a>', text)
-    return text
-
 
 # standard buttons with translatable labels
 BTN_OK, BTN_CANCEL, BTN_CLOSE, BTN_HELP, BTN_REFRESH = range(5)
@@ -400,6 +381,12 @@ class QBzrDialog(QtGui.QDialog, _QBzrWindowBase):
         
         self.windows = []
         self.closing = False
+        
+        # Even though this is a dialog, make it like a window. This allows us
+        # to have the best of both worlds, e.g. Default buttons from dialogs,
+        # and max and min buttons from window.
+        # It also fixes https://bugs.launchpad.net/qbzr/+bug/421039
+        self.setWindowFlags(QtCore.Qt.Window)
 
     def do_accept(self):
         self.accept()
@@ -407,6 +394,9 @@ class QBzrDialog(QtGui.QDialog, _QBzrWindowBase):
     def do_reject(self):
         self.reject()
 
+    def reject(self):
+        self.saveSize()
+        QtGui.QDialog.reject(self)
 
 throber_movie = None
 
@@ -510,95 +500,6 @@ def get_branch_config(branch):
         return get_global_config()
 
 
-def quote_tag(tag):
-    if _tag_re.search(tag):
-        return '"%s"' % tag
-    return tag
-
-
-def format_revision_html(rev, search_replace=None, show_timestamp=False):
-    props = []
-    if hasattr(rev, "revno"):
-        props.append((gettext("Revision:"), "%s revid:%s" % (rev.revno, rev.revision_id)))
-    else:
-        props.append((gettext("Revision:"), "revid:%s" % (rev.revision_id)))
-
-    def short_text(summary, length):
-        if len(summary) > length:
-            return summary[:length-1] + u"\u2026"
-        else:
-            return summary
-
-    def revision_list_html(revisions):
-        return ', '.join('<a href="qlog-revid:%s" title="%s">%s: %s</a>' % (
-            (r.revision_id,
-             htmlencode(get_summary(r)),
-             short_text(r.revno, 10),
-             htmlencode(short_text(get_summary(r), 60)))) for r in revisions)
-
-    parents = getattr(rev, 'parents', None)
-    if parents:
-        props.append((gettext("Parents:"), revision_list_html(parents)))
-
-    children = getattr(rev, 'children', None)
-    if children:
-        props.append((gettext("Children:"), revision_list_html(children)))
-
-    if not rev.revision_id == CURRENT_REVISION:
-        if show_timestamp and rev.timestamp is not None:
-            props.append((gettext("Date:"), format_timestamp(rev.timestamp)))
-    
-        props.append((gettext("Committer:"), htmlize(rev.committer)))
-        author = rev.properties.get('author')
-        if author:
-            props.append((gettext("Author:"), htmlize(author)))
-        else:
-            authors = rev.properties.get('authors')
-            if authors:
-                for author in authors.split('\n'):
-                    props.append((gettext("Author:"), htmlize(author)))
-    
-        branch_nick = rev.properties.get('branch-nick')
-        if branch_nick:
-            props.append((gettext("Branch:"), htmlize(branch_nick)))
-    
-        tags = getattr(rev, 'tags', None)
-        if tags:
-            tags = map(quote_tag, tags)
-            props.append((gettext("Tags:"), htmlencode(", ".join(tags))))
-    
-        bugs = []
-        for bug in rev.properties.get('bugs', '').split('\n'):
-            if bug:
-                url, status = bug.split(' ', 1)
-                bugs.append('<a href="%(url)s">%(url)s</a> %(status)s' % (
-                    dict(url=url, status=gettext(status))))
-        if bugs:
-            props.append((ngettext("Bug:", "Bugs:", len(bugs)), ", ".join(bugs)))
-
-    text = []
-    text.append('<table style="background:#EDEDED;" width="100%" cellspacing="0" cellpadding="0"><tr><td>')
-    text.append('<table cellspacing="0" cellpadding="0">')
-    for prop in props:
-        # <nobr> needed because in Russian some prop labels has 2 words
-        # &nbsp; needed because on Windows + PyQt 4.3.1 style=padding-left:5px does not working
-        text.append(('<tr><td style="padding-left:2px;" align="right"><b><nobr>%s </nobr></b></td>'
-            '<td>%s</td></tr>') % prop)
-    text.append('</table>')
-    text.append('</td></tr></table>')
-
-    if rev.revision_id == CURRENT_REVISION:
-        message = gettext("Uncommited Working Tree Changes")
-    else:
-        message = htmlize(get_message(rev))
-    if search_replace:
-        for search, replace in search_replace:
-            message = re.sub(search, replace, message)
-    text.append('<div style="margin:2px;margin-top:0.5em;">%s</div>' % message)
-
-    return "".join(text)
-
-
 def open_browser(url):
     try:
         import webbrowser
@@ -638,10 +539,6 @@ def format_timestamp(timestamp):
     return unicode(date.toString(QtCore.Qt.LocalDate))
 
 
-def htmlencode(string):
-    return string.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;").replace("\"", "&quot;")
-
-
 def is_valid_encoding(encoding):
     import codecs
     try:
@@ -669,12 +566,6 @@ def get_set_encoding(encoding, branch):
     return encoding
 
 
-class RevisionMessageBrowser(QtGui.QTextBrowser):
-
-    def setSource(self, uri):
-        pass
-
-
 def file_extension(path):
     """Return extension of the file.
     This function is smarter than standard os.path.splitext,
@@ -721,13 +612,13 @@ class FilterOptions(object):
     def to_str(self):
         s = []
         if self.deleted:
-            s.append(i18n.gettext('deleted files'))
+            s.append(gettext('deleted files'))
         if self.added:
-            s.append(i18n.gettext('added files'))
+            s.append(gettext('added files'))
         if self.renamed:
-            s.append(i18n.gettext('renamed files'))
+            s.append(gettext('renamed files'))
         if self.modified:
-            s.append(i18n.gettext('modified files'))
+            s.append(gettext('modified files'))
         return ', '.join(s)
 
     def check(self, status):
@@ -784,7 +675,7 @@ def iter_branch_related_locations(branch):
         if location is not None:
             yield url_for_display(location)
 
-# A helper to fill a 'pull' combo.
+# A helper to fill a 'pull' combo. Returns the default value.
 def fill_pull_combo(combo, branch):
     if branch is None:
         p = u''
@@ -793,6 +684,7 @@ def fill_pull_combo(combo, branch):
         p = url_for_display(branch.get_parent() or '')
         related = iter_branch_related_locations(branch)
     fill_combo_with(combo, p, related, iter_saved_pull_locations())
+    return p
 
 
 # A helper to fill a combo with values.  Example usage:
@@ -1078,3 +970,37 @@ def open_tree(directory, ui_mode=False,
             return None
         else:
             raise
+
+
+def launchpad_project_from_url(url):
+    """If url is a Launchpad code URL, get the project name.
+
+    @return: project name or None
+    """
+    # The format ought to be scheme://host/~user-id/project-name/branch-name/
+    from urlparse import urlsplit
+    scheme, host, path = urlsplit(url)[:3]
+    # Sanity check the host
+    if (host.find('bazaar.launchpad.net') >= 0 or
+        host.find('bazaar.launchpad.dev') >= 0):
+        parts = path.strip('/').split('/')
+        if len(parts) == 3 and parts[0].startswith('~'):
+            return parts[1]
+    return None
+
+
+def _shlex_split_unicode_linux(unicode_string):
+    """Split unicode string to list of unicode arguments."""
+    return [unicode(p,'utf8') for p in shlex.split(unicode_string.encode('utf-8'))]
+
+def _shlex_split_unicode_windows(unicode_string):
+    """Split unicode string to list of unicode arguments.
+    Take care about backslashes as valid path separators.
+    """
+    utf8_string = unicode_string.replace('\\', '\\\\').encode('utf-8')
+    return [unicode(p,'utf8') for p in shlex.split(utf8_string)]
+
+if MS_WINDOWS:
+    shlex_split_unicode = _shlex_split_unicode_windows
+else:
+    shlex_split_unicode = _shlex_split_unicode_linux
